@@ -426,7 +426,7 @@ int pgp_application_pgp_handler (BODY *m, STATE *s)
   pgperrfile = mutt_buffer_pool_get ();
   tmpfname = mutt_buffer_pool_get ();
 
-  fseeko (s->fpin, m->offset, 0);
+  fseeko (s->fpin, m->offset, SEEK_SET);
   last_pos = m->offset;
 
   for (bytes = m->length; bytes > 0;)
@@ -822,7 +822,7 @@ int pgp_verify_one (BODY *sigbdy, STATE *s, const char *tempfile)
     goto cleanup;
   }
 
-  fseeko (s->fpin, sigbdy->offset, 0);
+  fseeko (s->fpin, sigbdy->offset, SEEK_SET);
   mutt_copy_bytes (s->fpin, fp, sigbdy->length);
   safe_fclose (&fp);
 
@@ -936,7 +936,7 @@ BODY *pgp_decrypt_part (BODY *a, STATE *s, FILE *fpout, BODY *p)
   FILE *pgpin, *pgpout, *pgperr, *pgptmp;
   struct stat info;
   BODY *tattach = NULL;
-  int len;
+  size_t len;
   BUFFER *pgperrfile = NULL, *pgptmpfile = NULL;
   pid_t thepid;
   int rv;
@@ -964,7 +964,7 @@ BODY *pgp_decrypt_part (BODY *a, STATE *s, FILE *fpout, BODY *p)
    * the temporary file.
    */
 
-  fseeko (s->fpin, a->offset, 0);
+  fseeko (s->fpin, a->offset, SEEK_SET);
   mutt_copy_bytes (s->fpin, pgptmp, a->length);
   safe_fclose (&pgptmp);
 
@@ -1061,7 +1061,6 @@ int pgp_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
   STATE s;
   BODY *p = b;
   int need_decode = 0;
-  int saved_type;
   LOFF_T saved_offset;
   size_t saved_length;
   FILE *decoded_fp = NULL;
@@ -1089,7 +1088,6 @@ int pgp_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
 
   if (need_decode)
   {
-    saved_type = b->type;
     saved_offset = b->offset;
     saved_length = b->length;
 
@@ -1101,7 +1099,7 @@ int pgp_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
     }
     unlink (mutt_b2s (tempfile));
 
-    fseeko (s.fpin, b->offset, 0);
+    fseeko (s.fpin, b->offset, SEEK_SET);
     s.fpout = decoded_fp;
 
     mutt_decode_attachment (b, &s);
@@ -1129,7 +1127,6 @@ int pgp_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
 bail:
   if (need_decode)
   {
-    b->type = saved_type;
     b->length = saved_length;
     b->offset = saved_offset;
     safe_fclose (&decoded_fp);
@@ -1406,7 +1403,7 @@ char *pgp_findKeys (ADDRESS *adrlist, int oppenc_mode)
         if (! oppenc_mode && option(OPTCRYPTCONFIRMHOOK))
         {
           snprintf (buf, sizeof (buf), _("Use keyID = \"%s\" for %s?"), keyID, p->mailbox);
-          r = mutt_yesorno (buf, MUTT_YES);
+          r = mutt_query_boolean (OPTCRYPTCONFIRMHOOK, buf, MUTT_YES);
         }
         if (r == MUTT_YES)
         {
