@@ -69,9 +69,14 @@
 #include <assert.h>
 #endif
 
+/* For execvp environment setting in send_msg() */
+extern char **environ;
+
 extern char RFC822Specials[];
 
-const char MimeSpecials[] = "@.,;:<>[]\\\"()?/= \t";
+/* This corresponds to RFC2045 tspecials, along with space/tab.
+ * Note that it does not contain ".", which is in RFC822Specials */
+const char MimeSpecials[] = "@,;:<>[]\\\"()?/= \t";
 
 const char B64Chars[64] = {
   'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
@@ -81,6 +86,15 @@ const char B64Chars[64] = {
   '8', '9', '+', '/'
 };
 
+/* RFC4648 section 5 Base 64 Encoding with URL and Filename Safe Alphabet */
+const char B64Chars_urlsafe[64] = {
+  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
+  'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
+  'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
+  't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
+  '8', '9', '-', '_'
+};
+
 static void transform_to_7bit (BODY *a, FILE *fpin);
 
 static void encode_quoted (FGETCONV * fc, FILE *fout, int istext)
@@ -303,8 +317,8 @@ int mutt_write_mime_header (BODY *a, FILE *f)
   char buffer[STRING];
   char *t;
   char *fn;
-  int len;
-  int tmplen;
+  size_t len;
+  size_t tmplen;
 
   fprintf (f, "Content-Type: %s/%s", TYPE (a), a->subtype);
 
@@ -898,6 +912,8 @@ CONTENT *mutt_get_content_info (const char *fname, BODY *b)
   struct stat sb;
 
   if (b && !fname) fname = b->filename;
+  if (!fname)
+    return NULL;
 
   if (stat (fname, &sb) == -1)
   {
@@ -971,7 +987,7 @@ int mutt_lookup_mime_type (BODY *att, const char *path)
   char buf[LONG_STRING];
   char subtype[STRING], xtype[STRING];
   int count;
-  int szf, sze, cur_sze;
+  size_t szf, sze, cur_sze;
   int type;
 
   *subtype = '\0';
@@ -1102,7 +1118,7 @@ void mutt_message_to_7bit (BODY *a, FILE *fp)
     goto cleanup;
   }
 
-  fseeko (fpin, a->offset, 0);
+  fseeko (fpin, a->offset, SEEK_SET);
   a->parts = mutt_parse_messageRFC822 (fpin, a);
 
   transform_to_7bit (a->parts, fpin);
@@ -1378,8 +1394,8 @@ retry:
     else if ((WithCrypto & APPLICATION_SMIME) &&
              ((mutt_is_application_smime (hdr->content) & SMIMEENCRYPT) == SMIMEENCRYPT))
     {
-      chflags |= CH_MIME | CH_TXTPLAIN;
-      cmflags = MUTT_CM_DECODE | MUTT_CM_CHARCONV;
+      chflags |= CH_MIME | CH_NONEWLINE;
+      cmflags = MUTT_CM_DECODE_SMIME;
       pgp &= ~SMIMEENCRYPT;
     }
   }
@@ -1596,6 +1612,9 @@ BODY *mutt_make_file_attach (const char *path)
   BODY *att;
   CONTENT *info;
 
+  if (!(path && *path))
+    return NULL;
+
   att = mutt_new_body ();
   att->filename = safe_strdup (path);
 
@@ -1873,36 +1892,51 @@ static const char *find_word (const char *src)
 }
 
 /* like wcwidth(), but gets const char* not wchar_t* */
-static int my_width (const char *str, int col, int flags)
+static int my_width (const char *p, int col, int flags)
 {
   wchar_t wc;
   int l, w = 0, nl = 0;
-  const char *p = str;
+  size_t n, consumed;
+  mbstate_t mbstate;
 
-  while (p && *p)
+  if (!p)
+    return 0;
+
+  memset (&mbstate, 0, sizeof (mbstate));
+  n = mutt_strlen (p);
+
+  while (*p && n)
   {
-    if (mbtowc (&wc, p, MB_CUR_MAX) >= 0)
+    consumed = mbrtowc (&wc, p, n, &mbstate);
+    if (!consumed)
+      break;
+    if (consumed == (size_t)(-1) || consumed == (size_t)(-2))
     {
-      l = wcwidth (wc);
-      if (l < 0)
-	l = 1;
-      /* correctly calc tab stop, even for sending as the
-       * line should look pretty on the receiving end */
-      if (wc == L'\t' || (nl && wc == L' '))
-      {
-	nl = 0;
-	l = 8 - (col % 8);
-      }
-      /* track newlines for display-case: if we have a space
-       * after a newline, assume 8 spaces as for display we
-       * always tab-fold */
-      else if ((flags & CH_DISPLAY) && wc == '\n')
-	nl = 1;
+      if (consumed == (size_t)(-1))
+        memset (&mbstate, 0, sizeof (mbstate));
+      wc = replacement_char ();
+      consumed = (consumed == (size_t)(-1)) ? 1 : n;
     }
-    else
+
+    l = wcwidth (wc);
+    if (l < 0)
       l = 1;
+    /* correctly calc tab stop, even for sending as the
+     * line should look pretty on the receiving end */
+    if (wc == L'\t' || (nl && wc == L' '))
+    {
+      nl = 0;
+      l = 8 - (col % 8);
+    }
+    /* track newlines for display-case: if we have a space
+     * after a newline, assume 8 spaces as for display we
+     * always tab-fold */
+    else if ((flags & CH_DISPLAY) && wc == '\n')
+      nl = 1;
+
     w += l;
-    p++;
+    p += consumed;
+    n -= consumed;
   }
   return w;
 }
@@ -2098,7 +2132,7 @@ static int write_one_header (FILE *fp, int pfxw, int max, int wraplen,
   else
   {
     t = strchr (start, ':');
-    if (!t || t > end)
+    if (!t || t >= end)
     {
       dprint (1, (debugfile, "mwoh: warning: header not in "
 		  "'key: value' format!\n"));
@@ -2386,7 +2420,7 @@ int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, char *date,
   /* Add any user defined headers */
   for (; tmp; tmp = tmp->next)
   {
-    if ((p = strchr (tmp->data, ':')))
+    if ((p = strchr (NONULL (tmp->data), ':')))
     {
       q = p;
 
@@ -2434,7 +2468,7 @@ static void encode_headers (LIST *h)
 
   for (; h; h = h->next)
   {
-    if (!(p = strchr (h->data, ':')))
+    if (!(p = strchr (NONULL (h->data), ':')))
       continue;
 
     i = p - h->data;
@@ -2571,8 +2605,10 @@ send_msg (const char *path, char **args, const char *msg, char **tempfile)
 	  _exit (S_ERR);
       }
 
-      /* execvpe is a glibc extension */
-      /* execvpe (path, args, mutt_envlist ()); */
+      mutt_reset_child_signals ();
+
+      /* execvpe is a glibc extension, so just manually set environ */
+      environ = mutt_envlist ();
       execvp (path, args);
       _exit (S_ERR);
     }
@@ -2777,6 +2813,8 @@ mutt_invoke_sendmail (ADDRESS *from,	/* the sender */
 
   args[argslen++] = NULL;
 
+  i = send_msg (path, args, msg, option(OPTNOCURSES) ? NULL : &childout);
+
   /* Some user's $sendmail command uses gpg for password decryption,
    * and is set up to prompt using ncurses pinentry.  If we
    * mutt_endwin() it leaves other users staring at a blank screen.
@@ -2784,7 +2822,7 @@ mutt_invoke_sendmail (ADDRESS *from,	/* the sender */
   if (!option (OPTNOCURSES))
     mutt_need_hard_redraw ();
 
-  if ((i = send_msg (path, args, msg, option(OPTNOCURSES) ? NULL : &childout)) != (EX_OK & 0xff))
+  if (i != (EX_OK & 0xff))
   {
     if (i != S_BKG)
     {
@@ -2901,7 +2939,7 @@ static int _mutt_bounce_message (FILE *fp, HEADER *h, ADDRESS *to, const char *r
     if (!option (OPTBOUNCEDELIVERED))
       ch_flags |= CH_WEED_DELIVERED;
 
-    fseeko (fp, h->offset, 0);
+    fseeko (fp, h->offset, SEEK_SET);
     fprintf (f, "Resent-From: %s\n", resent_from);
 
     date = mutt_buffer_pool_get ();
@@ -3036,6 +3074,35 @@ ADDRESS *mutt_remove_duplicates (ADDRESS *addr)
   return (top);
 }
 
+/* Given a list of addresses, remove group label and end delimiter
+ * entries.  This is useful when generating an encryption key list,
+ * as we don't want to scan the label and empty delimiter entries. */
+ADDRESS *mutt_remove_adrlist_group_delimiters (ADDRESS *addr)
+{
+  ADDRESS *top = addr;
+  ADDRESS **last = &top;
+
+  while (addr)
+  {
+    if (addr->group || !addr->mailbox)
+    {
+      *last = addr->next;
+
+      addr->next = NULL;
+      rfc822_free_address(&addr);
+
+      addr = *last;
+    }
+    else
+    {
+      last = &addr->next;
+      addr = addr->next;
+    }
+  }
+
+  return (top);
+}
+
 static void set_noconv_flags (BODY *b, short flag)
 {
   for (; b; b = b->next)
@@ -3208,10 +3275,10 @@ int mutt_write_fcc (const char *path, SEND_CONTEXT *sctx, const char *msgid, int
      * this will happen, and it can cause problems parsing the mailbox
      * later.
      */
-    fseek (tempfp, -1, 2);
+    fseek (tempfp, -1, SEEK_END);
     if (fgetc (tempfp) != '\n')
     {
-      fseek (tempfp, 0, 2);
+      fseek (tempfp, 0, SEEK_END);
       fputc ('\n', tempfp);
     }
 
